/* heap.c,v 1.3.2.21 1993/05/21 19:11:34 ath Rel2_2 */

/* Implementation of a heap.  */
#include "repairs.h"

struct _heap_elt
{
  int key;
  int elt;
}; 

struct _heap
{
  int size;             /* Number of elements in the heap */
  struct _heap_elt elts[N+1];
  int iter;             /* Index to iterate over heap elements. */
}; 
struct _heap Heap;

/* Heap functions (or macros):
  heap_create()				Return a new heap.
  heap_insert(h, key, elt)		Insert an element in a heap.
  heap_min(h)				Return the minimum elt of heap.
  heap_extract_min(h)			Return and delete the min elt.
  heap_delete(h, elt)			Delete an element of the heap.
  heap_member(h, elt)			Tests for membership of elt in h.
*/

void
heap_create()
{
  Heap.size = 0;
  Heap.elts[0].key=unseen;
  Heap.elts[0].elt=NULLINT;
}

move(first, second)
int first, second;
{
  Heap.elts[second].key = Heap.elts[first].key;
  Heap.elts[second].elt = Heap.elts[first].elt;
}

/* Reheap after adding element to position k (on the bottom). */
upheap(k)
int k;
{
  struct _heap_elt spare;
  spare.key = Heap.elts[k].key;
  spare.elt = Heap.elts[k].elt;
  while (Heap.elts[k/2].key <= spare.key)  {
    move(k/2, k);
    k = k/2;
  }
  Heap.elts[k].key = spare.key;
  Heap.elts[k].elt = spare.elt;
}

/* Add element to the heap. */
void
heap_insert(key, elt)
     int key;
     int elt;
{
  Heap.elts[++Heap.size].key = key;
  Heap.elts[Heap.size].elt = elt;
  upheap(Heap.size);
}

/* Reheap after adding element to position k (on the top). */
downheap(k)
int k;
{
  int j;
  struct _heap_elt spare;
  spare.key = Heap.elts[k].key;
  spare.elt = Heap.elts[k].elt;
  while (k <= Heap.size/2) {
    j=k+k;
    if (j<Heap.size && Heap.elts[j].key<Heap.elts[j+1].key) j++;
    if (spare.key >= Heap.elts[j].key) break;
    move(j, k);  k=j;
  }
  Heap.elts[k].key = spare.key;
  Heap.elts[k].elt = spare.elt;
}

/* Find, delete, and return the smallest element of the heap. */
int
heap_extract_min()
{
  struct _heap_elt spare;
  spare.key = Heap.elts[1].key;
  spare.elt = Heap.elts[1].elt;
  move(Heap.size--, 1);
  downheap(1);
  return spare.elt;
}

/* Return the minimum elt of heap.   Do not delete */
int
heap_min()
{
  int answer; 
  if (Heap.size<1) answer=NULLINT;
  else answer=Heap.elts[1].elt;
  return(answer);
}

int
heap_find(elt)
int elt;
{
  int j, answer ; 
  j=1;
  while (Heap.elts[j].elt!=elt && j < Heap.size) {j++;}
  if (Heap.elts[j].elt==elt) answer=j;
  else answer=NULLINT;
  return (answer); 
}

/* update an element of the heap. */
heap_update(elt, key)
int elt, key;
{
  int k, old;
  k = heap_find(elt);
  old = Heap.elts[k].key;
  Heap.elts[k].key=key;
  if (key > old) upheap(k);
  else if (key < old) downheap(k);
}

