#include <math.h>
#include "repairs.h"

int
printclosestbad() 
{   
  int i, count=0, distance, node;
  distance = unseen;
  for (i=0; i<G; i+=1) 
    if (Member[i].wbstatus[errornum]==BAD && 
	  Member[i].disttosource[errornum]<distance) {
      distance = Member[i].disttosource[errornum];
      node = Member[i].wbnode;
    }
  printf ("Closest bad node: %d\n", node);
  return node;
}   

/************************************************************/
/* Hop-based */

/* neighborhood() returns the scope needed for a request from 
 * node "closest" to reach all of the bad nodes.
 * The scope is in terms of the length on each edge.
 */
int
neighborhood(closest)
int closest;
{
  int i, scope=0;	
  listpfs(closest);
  for (i=0; i<G; i+=1)
    if (Member[i].wbstatus[errornum]==BAD && 
	  pqdistance(Member[i].wbnode)>scope){
      scope = pqdistance(Member[i].wbnode);
    }
  printf ("Scope of bad neighborhood: %d\n", scope);
  return (scope);
}

int   
printclosestgood(closest)
{   
  int i, distance, node;
  distance = unseen;
  for (i=0; i<G; i+=1)
    if (Member[i].wbstatus[errornum]==GOOD && 
	pqdistance(Member[i].wbnode)<distance) {
      distance = pqdistance(Member[i].wbnode);
      node = Member[i].wbnode;
    }
  printf ("Closest good node: %d distance: %d\n", node, distance);
  localsource = node;
  return distance;
 
}

/* area returns the number of nodes reachable at scope "scope" from
 * node "closest".
 * This procedure is no longer called by anyone - replaced by ttl_area.
 */
int
area(closest, scope)
int closest, scope;
{
  int i, count=0;
  listpfs(localsource);
  for (i=0; i<G; i+=1)
    if (pqdistance(Member[i].wbnode) <= scope) ++count;
  if (verbose5==1) printval(scope);
  printf ("Local recovery from node: %d scope: %d\n", localsource, scope);
  printf ("Number of members reachable by local recovery: %d\n", count);
  return (count);
}

/************************************************************/
/* TTL-based, for links with a range of thresholds. */

/* ttl_neighborhood() returns the ttl needed for a request from 
 * node "closest" to reach all of the bad nodes.
 * The scope is in terms of the length on each edge.
 */
int
ttl_neighborhood(closest)
int closest;
{
  int i, newttl, ttl=0;	
  listpfs(closest);
  for (i=0; i<G; i+=1) {
    if (Member[i].wbstatus[errornum]==BAD) {
      newttl = pqttl(Member[i].wbnode);
      if (newttl>ttl) ttl = newttl;
    }
  }
  /* For simulations with threshold-defined neighborhoods only!!! 
  if (ttl>0&&ttl<31) 
    ttl=31;
  else if (ttl>31) 
    ttl=MAXTTL;
  */
  printf ("TTL of bad neighborhood: %d\n", ttl);
  return (ttl);
}

/* ttl_printclosestgood returns the min ttl needed to go from "closest"
    to a GOOD node */
int   
ttl_printclosestgood(closest)
{   
  int i, ttl, node;
  ttl = unseen;
  for (i=0; i<G; i+=1)
    if (Member[i].wbstatus[errornum]==GOOD && 
	pqttl(Member[i].wbnode)<ttl) {
      ttl = pqttl(Member[i].wbnode);
      node = Member[i].wbnode;
    }
  /* For simulations with threshold-defined neighborhoods only!!! 
  OLD: if (ttl<31) ttl=31;
  if (ttl<31) 
    ttl=31;
  else if (ttl>31) 
    ttl=MAXTTL;
   */
  printf ("Closest good node: %d ttl: %d\n", node, ttl);
  localsource = node;
  return ttl;
}

/* ttl_hops reports the hops needed to go from "source" to "dest". */
int 
hops_report(source, dest)
int source, dest;
{
  int hops;
  listpfs(source);
  hops = pqhops(dest);
  printf ("Hops from node %d to node %d are %d\n", source, dest, hops);
  return hops;
}

/* area returns the number of nodes reachable at ttl "ttl" from
 * node "source".
 */
int
ttl_area(source, ttl)
int source, ttl;
{
  int i, count=0;
  listpfs(source);
  for (i=0; i<G; i+=1)
    if (pqttl(Member[i].wbnode) <= ttl) ++count;
  /*
  if (verbose5==1) printval(ttl);
  */
  printf ("Local recovery from node: %d ttl: %d\n", source, ttl);
  printf ("Number of members reachable by local recovery: %d\n", count);
  return (count);
}

/* area2 returns the number of nodes reachable at ttl "ttl" from
 * nodes "source" and "source1".  Used for localtype=2.
 */
int
ttl_area2(source, source1, ttl)
int source, source1, ttl;
{
  int i, count=0;
  for (i=0; i<G; i+=1)
    Member[i].temp=0;
  listpfs(source);
  for (i=0; i<G; i+=1)
    if (pqttl(Member[i].wbnode) <= ttl) {
       Member[i].temp=1;
      ++count;
    }
  listpfs(source1);
  for (i=0; i<G; i+=1)
    if (pqttl(Member[i].wbnode) <= ttl &&
    	Member[i].temp == 0 ) 
      ++count;
  /*
  if (verbose5==1) printval(ttl);
  */
  printf ("Local recovery from node: %d ttl: %d\n", source, ttl);
  printf ("Number of members reachable by local recovery: %d\n", count);
  return (count);
}

/************************************************************/

