#include <math.h>
#include "repairs.h"

/* Nobadnodes() returns 1 if there are no bad nodes in the wb session,
 *  or if the bad edge has not yet been chosen.  */
int
nobadnodes(errornum)
int errornum;
{ 
  int i, answer, done;
  i=0; answer=1; done=0;
  if (badedge[errornum].first == -1) done=1;
  while (done==0 && answer==1 && i<G) {
    if (Member[i++].wbstatus[errornum]==BAD) answer=0;
  }
  return answer;
}

/* toomanybadnodes() returns 1 if there are too many bad nodes in the 
 *  wb session, or if the bad edge has not yet been chosen.  */
int
toomanybadnodes(errornum)
int errornum;
{ 
  int i, answer, done, count;
  i=0; answer=1; done=0; count=0;
  if (badedge[errornum].first == -1) done=1;
  else answer=0;
  while (done==0 && count<=G/10 && i<G) {
    if (Member[i++].wbstatus[errornum]==BAD) ++count;;
  }
  if (count>G/10) answer=1;
  return answer;
}

/* Markbadnodes marks nodes that are missing a packet. */
markbadnodes(errornum)
int errornum;
{
  int i, j, k, wbnum, good, label, sides[N];
  for (i=0; i<N; ++i) sides[i]=0;
  sides[badedge[errornum].first] = BAD; 
  good = badedge[errornum].second;
  while (sides[good]==0) {
    sides[good]=GOOD;
    good = pqparent(good);
  }
  for (i=0; i<N; ++i) {
    if (sides[i]==0) {
      j=i;
      while (sides[j]==0) j = pqparent(j);
      label = sides[j];
      j=i;
      while (sides[j]==0) {sides[j]=label; j = pqparent(j);};
    }
  }
  wbnum=0;
  for (i=0; i<N; ++i) {
    if (wb[i]==1) {
      for (j=wbnum; wbnum<j+ETHER; ++wbnum) {
        if (wbnum<G) {
	  Member[wbnum].wbstatus[errornum]=sides[i];
	  Member[wbnum].wbnode=i;
        }
      }
    }
  }
}

/* Picksession chooses the G members of the multicast group. */
picksession() 
{
  int dest, i, j, new, pickmember[N];
  float frac;
  if (G==N) {
    for (i=0; i<G; ++i) {
      new=i;
      wb[new]=1;
      Member[i].wbnode=new;
      wb_num[new]=i;
    }
  } else if (members==1 && N>G) {
      /* choose nodes 1 to G as members (excluding node 0) */
      for (i=0; i<G; ++i) {
        new=i+1;
        wb[new]=1;
        Member[i].wbnode=new;
        wb_num[new]=i;
    }
  } else if (members==2){
     /* choose members deterministically */
      for (i=0; i<G; ++i) {
        new=(int) (i*N/G);
        wb[new]=1;
        Member[i].wbnode=new;
        wb_num[new]=i;
    }

  }
  else {
    /* randomly choose session members */
    for (i=0; i<N; ++i) pickmember[i]=0;
    for (i=0; i<G; i+=ETHER) {
      /* pick the network node */
      frac = (float) random()/2147483647.0;
      new = ((int) (frac*N))%N; 
      while (pickmember[new]==1) 
        {++new; new = new%N;}
      pickmember[new]=1;
      wb_num[new]=i;
      wb[new]=1;
      for (j=i; j<i+ETHER; ++j) {
        if (j<G) Member[j].wbnode=new;
      }
    }
  }
}

/* Picksource chooses the source. */
picksource(errornum)
int errornum;
{
  if (sourcetype==0) {
    /* random source */
    float frac;
    frac = (float) random()/2147483647.0;
    source[errornum] = Member[ ((int) (frac*G))%G ].wbnode;
  } else if (sourcetype==-1) {
    source[errornum] = Member[0].wbnode;
  }
}

/* Pickfailure chooses the failed edge from the edges in the tree. 
 * badedge.second is a good node, and its child badedge.first is a bad
 * node.
 */
pickfailure(errornum)
int errornum;
{
  int num, degree = graphtype;
  float frac;
  struct node *t;
  frac = (float) random()/2147483647.0;
  if (edgetype==0) {
    /* random edge */
    badedge[errornum].first = ((int) (frac*N))%N;
    if (badedge[errornum].first == source[errornum]) 
	badedge[errornum].first = (source[errornum]+1)%N;
    badedge[errornum].second = pqparent(badedge[errornum].first);

  } else if (edgetype==-1) {
    /* random, edge adjacent to source */
    if (degree==0) 
      num=10; 
    else num = (int) (frac*(degree+1)) ;
    t = adj[source[errornum]];
    while (num>1 && t->next != z) {
      t = t->next;
      --num;
    }
    badedge[errornum].first = t->v;
    badedge[errornum].second = pqparent(badedge[errornum].first);
  } else if (edgetype==-2) {
    /* deterministic, random edge adjacent to source */
    t = adj[source[errornum]];
    badedge[errornum].first = t->v;
    badedge[errornum].second = pqparent(badedge[errornum].first);
  } else  if (edgetype==1) {
      /* edge at distance "edgetype" from source *
       * this assumes a tree, and that all nodes are session members */
      badedge[errornum].first = 1;
      badedge[errornum].second = pqparent(badedge[errornum].first);
  }  else  if (edgetype > 1 && degree > 2) {
      badedge[errornum].first = (int) (1 + 
	degree*(1 - power(degree-1, edgetype-1))/(2 - degree));
      badedge[errornum].second = pqparent(badedge[errornum].first);
      printf ("edge 1: %d\n", badedge[errornum].second);
      printf ("edge 2: %d\n", badedge[errornum].first); 
  }  else  if (edgetype > 1 && degree == 2) {
      badedge[errornum].first = 1 + 2*(edgetype-1);
      badedge[errornum].second = pqparent(badedge[errornum].first);
      printf ("edge 1: %d\n", badedge[errornum].second);
      printf ("edge 2: %d\n", badedge[errornum].first); 
  }
}
