/* Pq.c implements a priority queue.
 * TreeNode[i] contains the values associated with the entry i = pq[k]
 * in the tree.
 * The entries in TreeNode[].val are negative, and pq.c chooses the
 * largest entry.
 */
/* The priority queue will be implemented as a heap, sooner or later. */
  
#include <math.h>
#include "repairs.h"
  
struct treenode {
        int val;                /* val gives the distance from some
                                 *  source node */
        int hops;               /* the distance from that source node in hops */
	int ttl;		/* the TTL needed from the source to reach
				 *  this node. */
        int parent;             /* gives the parent node in the minimum
                                 *  spanning tree.  The root has parent  "-1".*/
};
struct treenode TreeNode[N+1];    /* This data structure is reused. */

 
pqinsert(v, metric, parent) 
int v, metric, parent;
{
  heap_insert(metric, v);
  TreeNode[v].val = metric;
  TreeNode[v].parent = parent;
  TreeNode[v].hops = TreeNode[parent].hops + 1;
  if (TreeNode[parent].ttl-TreeNode[parent].hops >= linkthresh(v, parent))
    TreeNode[v].ttl = TreeNode[parent].ttl;
  else {
    TreeNode[v].ttl = linkthresh(v, parent) + TreeNode[parent].hops;
  }
  /*
  printf ("parent: %d to: %d ttl: %d\n", parent, v, TreeNode[v].ttl);
  printf ("linkthresh: %d\n", linkthresh(v, parent));
  */
}     
      
pqinitialize(source)
int source;
{     
  int i;
  heap_create();
  heap_insert(0, source);
  for (i=0; i < N+1; i++) {
    TreeNode[i].val=-unseen;
    TreeNode[i].parent = -2;
    TreeNode[i].hops = 0;
    TreeNode[i].ttl = 0;
  }
  TreeNode[source].parent = -1;
  TreeNode[source].val = 0;
  TreeNode[source].hops = 0;
  TreeNode[source].ttl = 0;
}
  
/* Return distance from the source node. */
int
pqdistance(v)
int v;
{
  return TreeNode[v].val;
}
   
/* Return ttl needed from the source node. */
int
pqttl(v)
int v;
{
  return TreeNode[v].ttl;
}
   
/* Return parent node. */
int
pqparent(v)  
int v;
{ 
  return TreeNode[v].parent;
}     

/* Return hop count from the source node. */
int   
pqhops(v)
int v;
{ 
  return TreeNode[v].hops; 
}   
    
/* Pqupdate returns 1 if the node was not in the priority queue, or was
 * in the priority queue with a smaller (negative) metric.  Pqupdate
inserts
 * the node in the priority queue if necessary. */
pqupdate(v, metric, parent)
int v, metric, parent;
{ 
  int answer;
  answer=0;
  if (heap_find(v) == NULLINT) {
      answer=1;
      pqinsert(v, metric, parent);
  }
  else if (TreeNode[v].val < metric) {
      TreeNode[v].val = metric;
      answer=1;
      heap_update(v, metric);
  }
  return answer;
} 
  
/* Pqempty returns 1 if the priority queue is empty. */
int 
pqempty()
{
  int top;
  top = heap_min();
  if (top == NULLINT) return 1; else return 0;
} 
  
/* Pqremore removes the entry with the largest distance. */
/* For shortest path applications, these distances are negative. */
int   
pqremove()
{  
  int answer;
  answer = heap_extract_min();
  TreeNode[answer].val = -TreeNode[answer].val;
  return answer;
}     
