 /* Pq1.c implements a priority queue for timer values.
 * Timer[i] contains the values associated with the entry i = pq1[k]
 * in the tree.
 * The entries in timer[] are positive, and pq1.c chooses the smallest entry.
 *   Timer events for indices {0, G-1} are nacks for node i, and 
 * timer events for indices {G, 2G-1} are for repairs for node i%G. 
 */
/* This will be implemented as a heap at some point. */ 
  
#include "repairs.h"
 
int pq1top;
int type[2*G*G+1];		/* type, from eventstruct */
int wbnodeno[2*G*G+1];		/* node num, from 0 to G-1 */
int src[2*G*G+1];		/* source node for NACK, REPAIR */
float timer[2*G*G+1];		/* event time */
float dist[2*G*G+1];		/* distance used in setting timer */
  
pq1insert(type1, v, metric, distance, source)
int type1, v;
float metric, distance;
{ 
  ++pq1top;
  if (metric<0 || metric>MAXTIME) {
    abort();
  }
  if (pq1top>2*G*G+1) {
    printf ("Fatal error: pq1top in pq1insert: %d\n", pq1top);
    pq1print();
    abort();
  }
  type[pq1top]=type1;
  timer[pq1top]=metric;
  wbnodeno[pq1top]=v;
  dist[pq1top]=distance;
  src[pq1top]=source;
}
 
pq1initialize() 
{
  int i;
  pq1top = 0;
  for (i=0; i<G*G; i++) {
    timer[i]=(float) unseen;
    type[i]=0;
    wbnodeno[i]=-1;
    dist[i]=-1;
    src[i]=-1;
  }
}

pq1_reinit(min)
int min;
{
  timer[min]=(float) unseen;
  type[min]=0;
  wbnodeno[min]=-1;
  dist[min]=-1;
  src[min]=-1;
}
  
int
pq1empty()
{ 
  if (pq1top <= 0) { return 1; } 
  else return 0;
} 

/* Movetoptomin repacks the array after the removal of an entry. */
movetoptomin(pq1top, min)
int pq1top, min;
{
  wbnodeno[min] = wbnodeno[pq1top];
  type[min] = type[pq1top];
  timer[min] = timer[pq1top];
  dist[min] = dist[pq1top]; 
  src[min] = src[pq1top];
}
  
/* For a negative parameter, pq1remore removes the entry with the smallest 
 *  timer.  Otherwise pq1remove removes the entry specified by the
 *  parameter. */
pq1remove(v)
int v;
{
  int j, min, answer;
  if (pq1top<=0) abort();
  if (v<0) {
    min = 1;
    for (j=2; j <= pq1top; j++)
      if (timer[j] < timer[min]) min = j;
  }
  else if (v==0) abort();
  else min=v;
  event.wbnode = wbnodeno[min];
  event.type = type[min]; 
  event.time = timer[min];
  event.distance = dist[min];
  event.source = src[min];
  if (event.time<0) abort();  /* there was a negative timer entry */ 
  if (pq1top>1) {
    movetoptomin(pq1top--, min);
    pq1_reinit(pq1top+1);
  }
  else if (pq1top==1) {
    pq1top=0;
    pq1_reinit(1);
  }
  else pq1top=0;
} 

/* Pq1cancel cancels all timer entries for this node, and returns 1 if
 * in fact some timer was cancelled. */ 
int
pq1cancel(v)
int v;
{
  int j, answer=0;
  for (j=1; j <= pq1top; j++) 
    if (wbnodeno[j]==v&&(type[j]==RTIMER||type[j]==NTIMER)) {
      pq1remove(j);
      answer=1;
    }
  return answer;
}

/* if node v has an event of this type, replace it with an event
 *  with this metric. */
int
pq1update(type1, v, metric, distance)
int type1, v;
float metric, distance;
{
  int i, found=0;
  for (i=1; i <= pq1top; i++)
    if (type[i]==type1 && wbnodeno[i]==v) {
      if (metric<0 || metric > MAXTIME) abort();
      timer[i] = metric; dist[i] = distance;
      found=1;
    }
  return found;
}

/* Pq1print prints all of the entries in the event queue. */
pq1print()
{
  int i;
  printf ("Event queue: \n");
  for (i=1; i<=pq1top; i++) {
    printf ("  node: %d time: %6.3f type: ",
      Member[wbnodeno[i]].wbnode, timer[i]);
  printtype(type[i]);
  }
  printf ("\n");
}

/* if this node has another event of this type, then this is ignored.
 * Otherwise, this is inserted.  Returns "1" if a new insertion was done.
 */
int
pq1check(type1, v, metric, distance)
int type1, v;
float metric, distance;
{
  int i, found;
  found=0;
  i=1;
  while (found==0&&i<= pq1top) {
    if (type[i]==type1 && wbnodeno[i]==v) found=1;
    else ++i;
  }
  if (found==0) {
    pq1insert(type1, v, metric, distance, -1);
  }
  return !found;
}
