/* make, make install */
/* ./repairs < stand.in */
/* distance:
  bad nodes need to know distance from source of error.
    For first setting NTIMER, bad nodes need to know the distance
      from the source.
    For backing off on NTIMER, bad nodes need to know the original
    distance of the timer.
  good nodes need to know distance from repair request.
    For backing off on RTIMER, good nodes need to know the original
    distance from the NACK source.  Bad nodes don't need to know the
    distance from the NACKk source.
*/

#include <math.h>
#include "repairs.h"

main ()
{
  int errornum, k, closest, badttl, goodttl, scope, reach, ttl,
    oldseed;
  initialize();
  makegraphlst();		/* Randomly generate the graph. */
  srandom(firstseed);
  if (verbose5==1) printgraph();
  picksession();		/* Randomly choose wb members.  */
  if (verbose5==1) printgroup();
  for (errornum=0; errornum<ERRORS; ++errornum) {
    picksource(errornum);		/* Randomly choose source. */
    if (verify==1) verifysession(errornum);
    if (verbose5==1) printsource(errornum);
    listpfs(source[errornum]);	/* Make shortest path tree from source.  */
    if (verbose2==1) printval(errornum);
    /* printwb(errornum); */
    while (nobadnodes(errornum)==1 || 
         (simtype==2 && toomanybadnodes(errornum)==1)) {	
         /* Try again if there were no bad nodes, or too many. */ 
      pickfailure(errornum);		/* Pick failed edge. */
      markbadnodes(errornum);		/* Mark nodes to set nack timers. */
      if (verbose5==1) {
        printf ("Choosing failed edge:\n");
        printbadedge(errornum);
      }
      printbadnodes(errornum);
    }
    if (verbose7==1) {
      printgraph();
    }
    for (k=0; k<G; k++) 
      if (Member[k].wbstatus[errornum]==BAD) 
        Member[k].disttosource[errornum] = (float) pqdistance(Member[k].wbnode);
    if (verbose3==1) printdisttosource(errornum); 
  }
  if (seed1 == -1) (void) time(&seed1);
  if (seed1 != 0) srandom(seed1);
  if (simtype==0) for (round=0; round<repetitions; round++) {
    errornum = (round/TURNS)%ERRORS; 
    if (rep_to_save==0||rep_to_save==round) printf ("Round: %d\n", round);
    if (round > 0) reinit();
    setnacktimers();		/* Set nack timers. */
    if (verbose==1) printnacktimes();
    processevents();   		/* Process nack timers. */
  }
  if (simtype==1||simtype==2) {
    /* local recovery stats */
    /* To do:  For each bad node, compute TTL needed to get there from
     * source of request.  For the "closest" good node, compute TTL
     * needed to get there from source of request.
     * For localtype 1, TTL from repair is TTL of request + hops to requestor.
     * For localtype 2, TTL of first repair is TTL to requestor, and
     *    TTL of second repair is TTL of original request.
     */
    closest = printclosestbad();
    badttl = ttl_neighborhood(closest);
    goodttl = ttl_printclosestgood(closest); 
    ttl = badttl;  if (goodttl>ttl) ttl=goodttl;
    if (localtype==1) {
      ttl += hops_report(localsource, closest);
      reach = ttl_area(localsource, ttl);
    }
    else if (localtype==2) {
      reach = ttl_area2(localsource, closest, ttl);
    }
  /*
  print closest bad node, max diameter "badttl" for bad area.
  find closest good node, distance "goodttl" from bad node.
  Request has ttl "ttl = Max(badttl, goodttl)".
  For localtype 1, repair has ttl "Max(badttl, goodttl) + hops to request".  
  For localtype 2, the repair is sent twice, first from goodnode
    with ttl "Max(badttl, goodttl)", and second from badnode with
    the same ttl.
    Find total number of nodes reached.
  */
  }
  if (simtype==3) {
    /* request delay for closest node */
    setnacktimers();
    processevents();
  }
}

repeat()
{
  printf ("Second time:\n");
}


/* Addevents adds new REPAIR or NACK events after a timer expires. */
addevents(newsource, ether)
int newsource, ether;
{
  int i, newtype;
  float time, dist;
  for (i=0; i<G; ++i) {
    if (Member[i].wbnode != newsource || i%ETHER != ether) {
      if (event.type==RTIMER) newtype=REPAIR;
      if (event.type==NTIMER) newtype=NACK;
      dist = (float) pqdistance(Member[i].wbnode);
      if (Member[i].wbnode == newsource) 
        time = event.time + dist + DELAY;
      else 
	time = event.time + dist + 2*DELAY;
      if (time<0) abort();
      pq1insert(newtype, i, time, (float) pqdistance(Member[i].wbnode), newsource);
    }
  if (verbose1==1) pq1print();
  }
}

/* Compute the average request delay, after first receiving or sending *
 * a NACK. */
make_avg_req(nodeno, delay)
int nodeno;
float delay;
{
  if (delay>Member[nodeno].maxreqdelay)
    Member[nodeno].maxreqdelay=0;
  if (Member[nodeno].avereqdelay==-1)
	  Member[nodeno].avereqdelay=delay;  
  else Member[nodeno].avereqdelay=(1-WEIGHT)*Member[nodeno].avereqdelay
 	 + WEIGHT*delay;
  Member[nodeno].inittime[errornum]=0;
}

/* Compute the average repair delay */
make_avg_rep(nodeno, delay)
int nodeno;
float delay;
{
  if (delay>Member[nodeno].maxrepdelay)
    Member[nodeno].maxrepdelay=0;
  if (Member[nodeno].averepdelay==-1)
	  Member[nodeno].averepdelay=delay;  
  else Member[nodeno].averepdelay=(1-WEIGHT)*Member[nodeno].averepdelay
	+ WEIGHT*delay;
  Member[nodeno].inittime[errornum]=0;
}

/* Processtimer() backs off for NTIMERs, multicasts nacks or repairs. */
processtimer()
{
  int nodeno, newsource, ether;
  float delay, dist;
  float now=event.time;
  nodeno=event.wbnode;
  if (event.type==NTIMER) {
    dist=Member[nodeno].disttosource[errornum];
    ++Member[nodeno].dupreqs;
    delay=(float) (now - Member[nodeno].inittime[errornum])/
          (2.0*dist);
    if (now > Member[nodeno].backofftime[errornum])
      timerbackoff(nodeno, dist, event.type);
    if (Member[nodeno].inittime[errornum]>0&&
      Member[nodeno].wbstatus[errornum]==BAD ) {
      delaystats(nodeno, delay, dist);
      make_avg_req(nodeno, delay);
    }
    initreq(nodeno);
  }
  else if (event.type==RTIMER) {
    ++Member[nodeno].dupreps;
    Member[nodeno].ignorenack=now+3*Member[nodeno].disttosource[errornum];
    if (rep_to_save==0||rep_to_save==round) {
     printf ("Action: node %d multicast REPAIRs, ignores NACKs til %6.3f",
	  Member[nodeno].wbnode, Member[nodeno].ignorenack);
     if (ETHER > 1) printf (" ether: %d", nodeno%ETHER);
     printf ("\n");
    }
    if (Member[nodeno].inittime[errornum]>0 &&
      Member[nodeno].wbstatus[errornum]==GOOD) {
      delay=(float) (now - Member[nodeno].inittime[errornum])/
        (2.0*Member[nodeno].disttosource[errornum]);
      make_avg_rep(nodeno, delay);
    }
    initrep(nodeno);
  }
  else abort();
  newsource=Member[nodeno].wbnode;  	/* send messages */
  ether=nodeno%ETHER;
  listpfs(newsource);
  if (verbose2==1) printval(0); 
  addevents(newsource, ether);
}

/* Setrepair sets the time for the repair timer.
 * For a distance d from the sender, this time is uniformly chosen
 * from [RSTART d, (RSTART + RSTOP) d], for a session with G members.
 */
float
setrepair1(v, distance)
int v;
float distance;
{
  float frac, time1, mintime, maxtime;
  frac = (float) (random()/2147483647.0);
  time1 = (RSTART + Member[v].rstartplus) * distance 
	+ (RSTOP + Member[v].rstopplus) * frac * distance;
  mintime = (RSTART + Member[v].rstartplus) * distance;
  maxtime=mintime + (RSTOP + Member[v].rstopplus) * distance;
  if (rep_to_save==0||rep_to_save==round)
   printf ("Info: node %d dist: %6.3f repair interval %6.3f %6.3f num: %6.3f ether: %d v: %d num: %6.3f\n",
        Member[v].wbnode, Member[v].disttosource[errornum], 
	mintime, maxtime, time1, v%ETHER, v, Member[v].rstartplus);
  return time1;
}

/* Compute the request delay stats */
delaystats(nodeno, delay, dist)
int nodeno;
float delay, dist;
{
  if (delay < reqdelay && dist <= reqdist ) {
    reqdelay = delay;
    reqdist = dist;
    printf ("ReqDelay: node: %d distance %6.3f request delay: %6.3f\n",
      nodeno+1, dist, delay);
  }
}

/* Processnack() backs off NTIMER for a "bad" node. 
 * A bad node also computes reqdelay and avereqdelay.
 * A good node sets RTIMER and saves the time.
 */
processnack()
{
  int nodeno, node1;
  float frac, now, time, dist1, delay, inittime;
  nodeno=event.wbnode;
  now=event.time;
  dist1 = Member[nodeno].disttosource[errornum];
  inittime = Member[nodeno].inittime[errornum];
  if (Member[nodeno].wbstatus[errornum]==BAD ||
	Member[nodeno].wbstatus[errornum]==WASBAD) {
    ++Member[nodeno].dupreqs;
    if (Member[nodeno].closest_req==1 && event.distance < 1.5 * dist1 )
      Member[nodeno].closest_req=0;
  }
  if (Member[nodeno].wbstatus[errornum]==BAD) { 		
    if (simtype==0 && now > Member[nodeno].backofftime[errornum])
      { timerbackoff(nodeno, dist1, NTIMER); }
    else if (simtype==3) { timercancel(nodeno, dist1, NTIMER); }
    if (inittime>0) {
      delay=(float) (now - inittime)/ (2.0*dist1);
      delaystats(nodeno, delay, dist1);
      make_avg_req(nodeno, delay);
    }
  }
  if (simtype==0 && Member[nodeno].wbstatus[errornum]!=BAD && 
		now > Member[nodeno].ignorenack) {	
    if (dist1<0) {
      Member[nodeno].disttosource[errornum] = event.distance;
      dist1 = Member[nodeno].disttosource[errornum];
    }
    time = now + setrepair1(nodeno, dist1);
    if (pq1check(RTIMER, nodeno, time, dist1)==1) {
     if (rep_to_save==0||rep_to_save==round) {
      printf 
        ("Action: time %6.3f distance %6.3f node %d set RTIMER for %6.3f", 
         event.time, dist1, Member[nodeno].wbnode, time);
      if (ETHER > 1) printf (" ether: %d", nodeno%ETHER);  
      printf ("\n"); 
     }
     Member[nodeno].inittime[errornum] = event.time;
    }
  }
}

/* Processrepair() cancels ourstanding repair or nack timers. 
 * A Good node computes "repdelay" and "averepdelay".
 */
processrepair()
{
  int nodeno;
  float delay, dist;
  float now=event.time;
  nodeno=event.wbnode;
  dist = Member[nodeno].disttosource[errornum];
  Member[nodeno].ignorenack = now + 
	3 * dist;
  if (rep_to_save==0||rep_to_save==round) {
    printf ("Action: node %d received REPAIR, ignores NACKs til %6.3f",
		  Member[nodeno].wbnode, Member[nodeno].ignorenack);
    if (ETHER > 1) printf (" ether: %d", nodeno%ETHER);
    printf ("\n"); 
  }
  if (pq1cancel(nodeno)==1) {
    if (rep_to_save==0||rep_to_save==round) 
      printf ("Action: node %d cancelled ", Member[nodeno].wbnode);
    if (Member[nodeno].backoff[errornum]<5) 
      Member[nodeno].backoff[errornum] += 1;
    if (rep_to_save==0||rep_to_save==round) {
      printresponse(nodeno);
      printf (" for time %6.3f distance %6.3f\n", 
	  event.time, Member[nodeno].disttosource[errornum]); 
    }
  }
  if (Member[nodeno].wbstatus[errornum]==BAD) 
    Member[nodeno].wbstatus[errornum]=WASBAD;
  if (Member[nodeno].wbstatus[errornum]==GOOD){
    ++Member[nodeno].dupreps;
    if (Member[nodeno].closest_rep==1 && event.distance < 1.5 * dist )
      Member[nodeno].closest_rep=0;
    if (Member[nodeno].inittime[errornum]>0) {
      delay=(float) (now - Member[nodeno].inittime[errornum])/
        (2.0*dist);
      make_avg_rep(nodeno, delay);
    }
  }
}

/* In detectloss(), node "node" detects a missing packet, sets a NTIMER,
 * and records the time.
 */
detectloss(node)
int node;
{
  pq1insert(NTIMER, node, Member[node].wbtimer, 0.0, -1); 
  Member[node].inittime[errornum]=Member[node].disttosource[errornum];
  if (Member[node].backoff[errornum]<5) 
      Member[node].backoff[errornum] += 1;
  /* Add dropid to list. */ 
  /* Upate constants. */
}

processevents()
{
  int i;
  float time;
  pq1initialize();
  for (i=0; i<G; ++i)
    if (Member[i].wbstatus[errornum]==BAD) {
      detectloss(i);
    }
  if (verbose1==1) pq1print();
  while (pq1empty()==0) {
    pq1remove(-1);
    now = event.time;
    if (now < 0) abort();
    i=event.wbnode;
    if (verbose==1 && (rep_to_save==0||rep_to_save==round)) {
      printf ("Event: node: %d time: %6.3f type: ", 
	Member[i].wbnode, event.time);
      printtype(event.type);
      if (event.type==NACK||event.type==REPAIR)
	printf (" from node: %d dist: %6.3f", event.source, event.distance);
      if (ETHER > 1) printf (" ether: %d", i%ETHER);
      printf ("\n");
    }
    if ((simtype==0 && event.type==RTIMER) || event.type==NTIMER) 
      processtimer(); 
    if (event.type==NACK) 
      processnack();
    if (event.type==REPAIR) 
      if (simtype==0) processrepair();

  }
}

power(x,n)
int  x,n;
{  int i, p;
   p=1;
   for ( i = 1; i <= n; ++i)
     p = p * x;
   return (p);
} 
