/*
To compile:  cc repairs.c -lm -g -o repairs
*/

#define srandom(seed) srand48((long)seed)
#define random() lrand48()
/* To use srandom and random with gcc on owl */

#define NULLINT -1
#define EPSILON 0.1
#define N 10                    /* number of nodes */
#define M 12			/* number of edges */
#define G 6			/* number of nodes in multicast group */
#define ETHER 1			/* number of ethernet nodes attached to
				 * each network node */
#define DELAY 0			/* delay on ethernets */
#define ERRORS 1		/* number of source/failed link pairs */
#define TURNS 1                 /* number of consecutive rounds/error */
#define TRUE    1
#define FALSE   0
#define WEIGHT 0.25		/* weight for averaging, usually 0.25 */ 
#define CLOUD 2	        	/* for structured trees with large
				 * thresholds, the number of rows in tree */
				/* CLOUD=2 and degree=4 gives 17 nodes/cloud */
#define THRESH 1		/* threshold between clouds */
#define THRESH2 1		/* threshold between second-level clouds */
#define ROW 3			/* rows in a mesh-based tree. */
#define DUPS 0			/* 1 for duplicates-based version,
				 * 0 for delay-based version */
#define MAXTTL 1000
#define MAXTIME 100000000	/* to protect against bugs */
/* NSTART, NSTOP for request timer.  RSTART, RSTOP for repair timer. */
#define NSTART 2		/* interval for setting request timer *
				 * minimum of 1, for backing off timers */
#define NSTOP 2
#define RSTART (log10((double)G)) 
#define RSTOP (log10((double)G))
/* (log10((double)G)) */
#define MINNSTART 0.5
#define MAXNSTART 2 
#define MINNSTOP 1
#define MAXNSTOP G
/* (sqrt((double)G)) */
/*
#define MINRSTART 0.5
*/
#define MINRSTART 0.5
#define MAXRSTART 2 
#define MINRSTOP 1
#define MAXRSTOP G

#define inserttype 3		/* -1 to randomly add extra edges */
				/* k to add edges across hop distance k */
#define localtype 1		/* 1 for TTL-based local recovery,
				   2 for TTL-based local recovery with
				     three-step recovery */
#define unseen 10000000
#define verbose 1		/* to print event messages */
#define verbose1 0		/* to print the event queue */
#define verbose2 0		/* to print shortest path trees */
#define verbose3 0		/* print disttosource */
#define verbose4 0		/* to check the graph construction */
#define verbose5 1		/* to print graph */
#define verbose6 0              /* to check random tree */
#define verbose7 0		/* to print graph */
#define verbose8 0		/* to check avedelay, avedups */
#define verbose9 0		/* to print adaptive variables */
#define verbose10 0		/* to print messages for simtype 3 */
#define verify 1
#define COLUMNS 10

#define REPAIR 	0
#define NACK	1
#define RTIMER	2
#define NTIMER	3

#define GOOD 1
#define BAD 2
#define WASBAD 3

long random();

int seed,			/* seed for creating graph.  
				 *	"0" to pick random */
    firstseed,			/* the initial seed actually used */
    seed1,			/* seed for setting timers. 
				 *	"0" to continue as is 
				 *	"-1" to select a new value */
    repetitions,		/* number of times to repeat experiment. */
    adapttimers,		/* "1" for yes, "0" for no */
    edgetype,			/* "0" for random, "-1" for near source,
				 *   "k" (k>0) for near source of distance 
				 *   at most k */
    sourcetype,			/* "0" for random, "-1" for node 0 */
    simtype,			/* "0" for repairs, 
				 * "1", "2" for local recovery stats */
    lengthtype,			/* "0" for all lengths=1,
				 * "1" for lengths from 1 to 10 */
    graphtype,			/* "0" for random tree,
				 * "k" for bounded tree */
    rep_to_save,		/* 0 to save all repetitions,
				 *   k>0 to save only the kth */
    threshtype,			/* 1 for an unstructured tree, 
				 *   all thresholds to be 1
				 * 2 for a structured tree.  */
    members,			/* 0 to choose randomly,
				 * 1 to choose 1-G (excluding node 0) */
    round,			/* if multiple repetitions */
    errornum,			/* if multiple repetitions */ 
    source[ERRORS],		/* the source node */
    localsource,                /* the repair source node for local recovery */
    cloudnum,			/* the number of nodes in a cloud */
    cloudnum1,			/* the number of nodes in a mesh cloud */
    dropid,			/* the ID of the current dropped packet */
    wb[N],			/* "1" index for nodes in wb session. 
				 * Used only for initialization. */
    wb_num[N],			/* session number of network node (1 to G) */
    MaxDups;			/* for adaptive algorithm */

float
    now,			/* current time */
    reqdelay,			/* for reporting request delay */
    reqdist,			/* distance, for reporting request delay */
    MaxDelay,			/* for adaptive algorithm */
    MaxAveDelay,		/* for adaptive algorithm */
    MaxAveDups;			/* for adaptive algorithm */

struct node { 
    int v; 			/* the net ID of the node (1 to N) */
    int member_num;		/* the member number of the node (1 to G) */
    int length;			/* distance in seconds */
    int thresh;			/* threshold */
    struct node *next;
};
struct node *z;
struct node *adj[N];		/* the adjacency list for the graph */

struct edge { 
	int first; 
	int second;
	int length;
}; 
struct edge badedge[ERRORS];	/* the failed edge */
struct edge addedge;		/* the next edge to add */

struct wbmember {
	int wbnode;             /* the net IDs of nodes in the wb session */
        int wbstatus[ERRORS];	/* GOOD for a node that didn't miss packet #1, 
				 * BAD for a node that did miss packet #1 */
        int ttl[ERRORS];   	/* For BAD nodes, TTL to get here from
				 *  source of request. */
    	int backoff[ERRORS];	/* 0 for first time timer is set, 
				 *  increments by one for each backoff, 
				 *  up to 6  */
	int dupreqs;		/* count of duplicate requests */
	int dupreps;		/* count of duplicate repairs */
	int maxdupreqs;		/* max duplicate requests */
	int maxdupreps;		/* max duplicate repairs */
	int closest_req;        /* "1" if request sent, any dup requests 
				 *  are from further away */
	int closest_rep;        /* "1" if repair sent, any dup repairs 
				 *  are from further away */
	int temp;               /* for temporary state */
	float avedupreqs;	/* average duplicate requests */
	float avedupreps;	/* average duplicate repairs */
	float maxreqdelay;	/* request delay/RTT.  For BAD nodes. Time from
				 * setting of request time until receipt or
				 * sending of request/RTT to source.  */
	float maxrepdelay;	/* repair delay/RTT.  For GOOD nodes. Time from
				 * setting of repair timer until receipt or 
				 * sending of repair/RTT to requestor.  */ 
        float avereqdelay;	/* average request delay */
        float averepdelay;	/* average repair delay */
    	float nstartplus;       /* constant to add to NSTART */
    	float nstopplus;	/* constant to add to NSTOP */
    	float rstartplus;	/* constant to add to RSTART */
    	float rstopplus;	/* constant to add to RSTOP */
    	float wbtimer;		/* expiration time for timer */
    	float disttosource[ERRORS];	
				/* For BAD nodes, distance from source.
				 * For GOOD nodes, the distance from 
				 * the source of the first NACK received.*/
        float inittime[ERRORS];
				/* For BAD nodes, time request timer set.
				 * For GOOD nodes, time repair timer set.*/
	float backofftime[ERRORS];
				/* Don't backoff timer unless current
				 *  time is greater than this time.
				 *  Initialized to 0. */
    	float ignorenack;	/* ignore arriving nacks until ignorenack[i] */

};
struct wbmember Member[G];	/* members of the wb session */

struct eventstruct { 
    int wbnode; 		/* wb node */
    int type; 			/* REPAIR for repair receipt,
				 * NACK for nack receipt, 
				 * RTIMER for repair timer expiration,
				 * NTIMER for nack timer expiration.
				 */
    int source;			/* source of NACK, REPAIR */
    float time;			/* time of event */
    float distance;		/* used in setting timer */
};
struct eventstruct event;	/* event to be processed, in order. */
