#include <math.h>
#include "repairs.h"

/* Setnack sets the time for the repair request timer.
 * For a distance d from the source, this time is uniformly chosen
 * from [NSTART d, (NSTART + NSTOP) d].
 */
float
setnack(v, distance, backoffnum)
int v, backoffnum;
float distance;
{
  int nodeno, factor;
  float frac, time1, mintime, maxtime, now;
  frac = (float) random()/2147483647.0;
  nodeno = v;
  time1 = (NSTART + Member[nodeno].nstartplus) * distance 
	+ (NSTOP + Member[nodeno].nstopplus) * frac * distance;
  factor = power(3, backoffnum);
  time1 *= factor;
  mintime = Member[v].disttosource[errornum] + 
	factor*(NSTART+Member[nodeno].nstartplus)*distance;
  maxtime=mintime + factor*(NSTOP + Member[nodeno].nstopplus)*distance;
  if (rep_to_save==0||rep_to_save==round)
   printf ("Info: node %d dist: %6.3f req. int.: %6.3f %6.3f num: %6.3f ether: %d nstartplus: %6.3f\n",
	Member[v].wbnode, Member[v].disttosource[errornum], 
	mintime, maxtime, time1, v%ETHER, Member[nodeno].nstartplus);
  return time1;
} 
  
/* Timerbackoff backs off on NTIMER. */
/* I need to add different timers for different errors. */
timerbackoff(v, dist, type)
int v, type;
float dist;
{
  int found;
  float time;
  time = event.time + setnack(v, dist, Member[v].backoff[errornum]);
  if (simtype==0) {
    found = pq1update(type, v, time, dist);
    if (found==0) pq1insert(type, v, time, 0.0, -1);
  }
  else if (simtype==3) {
    found = pq1cancel(v);
    time = 0.0;
  }
  if (rep_to_save==0||rep_to_save==round) {
    printf ("Action: distance %6.3f node %d backed off ", 
      dist, Member[v].wbnode);
    printtype(type);
    printf (" to %6.3f", time);
    if (found==0) {
      printf (", multicast ");
      printpackets(type);
    }
    if (ETHER > 1) printf (" ether: %d", v%ETHER);
    printf ("\n");
  }
  if (Member[v].backoff[errornum]<5)
	Member[v].backoff[errornum] += 1;
}

/* Timercancel cancels NTIMER. */
timercancel(v, dist, type)
int v, type;
float dist;
{
  int found;
  pq1cancel(v);
  if (verbose10==1) 
    printf ("Action: distance %6.3f node %d cancelled timers\n", 
      dist, Member[v].wbnode);
}

/* Setnacktimers sets the timer expiration value for each bad node's
 * nack timer.  This uses the wbnode, wbstatus, and val arrays.
 * Assume that the source sends the first packet after the dropped
 * packet at time 0.
 */
setnacktimers()
{
  int i, nodeno;
  float now, distance;
  for (i=0; i<G; ++i) {
    if (Member[i].wbstatus[errornum]==BAD) {
      nodeno = Member[i].wbnode;
      distance = Member[i].disttosource[errornum];
      if (distance < reqdist) reqdist = distance;
      now = Member[i].disttosource[errornum];
      Member[i].wbtimer = now + 
	setnack(i, distance, Member[i].backoff[errornum]);
      Member[i].backofftime[errornum]=(Member[i].wbtimer-now)/2.0;
    }
  }
}
